from __future__ import annotations

import asyncio
import typing as ty
from abc import ABC, abstractmethod
from functools import wraps

from ..exceptions import ConnectionIsNotAccrued
from .base import BaseConnectionWrapper, BasePoolConnectionWrapper, BaseProvider

if ty.TYPE_CHECKING:
    from .base import InsertQuery, Query, SelectQuery


def translate_exceptions(func):
    @wraps(func)
    async def _wrapper(self: BaseAsyncProvider, query, args=()):
        try:
            return await func(self, query, args)
        except Exception as e:
            raise self._translate_exception(e, query, args)

    return _wrapper


class BaseAsyncProvider[ConnType](BaseProvider[ConnType], ABC):
    """
    Base class for async SQL providers.
    ConnType : type of connection instance.
    """

    @abstractmethod
    def ensure_connection(
        self,
    ) -> (
        AsyncConnectionWrapper[ConnType] | AsyncPoolConnectionWrapper[ConnType]
    ):
        raise NotImplementedError()

    @abstractmethod
    async def _fetchone(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> tuple[ty.Any, ...] | None:
        raise NotImplementedError()

    @abstractmethod
    async def _fetchall(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> list[tuple[ty.Any, ...]]:
        raise NotImplementedError()

    @translate_exceptions
    async def execute(self, query: Query, args: ty.Sequence[ty.Any] = ()):
        args = tuple(self.adapt_value(arg) for arg in args)
        return await self._execute(query, args)

    @translate_exceptions
    async def execute_insert_query(
        self, query: InsertQuery, args: ty.Sequence[ty.Any]
    ) -> list[int]:
        args = tuple(self.adapt_value(arg) for arg in args)
        return [row[0] for row in await self._fetchall(query, args)]

    @translate_exceptions
    async def fetchone(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> tuple[ty.Any, ...] | None:
        args = tuple(self.adapt_value(arg) for arg in args)
        return await self._fetchone(query, args)

    @translate_exceptions
    async def fetchall(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> list[tuple[ty.Any, ...]]:
        args = tuple(self.adapt_value(arg) for arg in args)
        return await self._fetchall(query, args)


class AsyncConnectionWrapper[ConnType](BaseConnectionWrapper[ConnType]):
    def __init__(self, provider: BaseProvider[ConnType], lock: asyncio.Lock):
        self.provider = provider
        self.connection: ConnType | None = None
        self._lock = lock

    async def ensure_connection(self) -> None:
        if not self.connection:
            await self.provider.create_connection()
            self.connection = self.provider.connection

    async def __aenter__(self) -> ConnType:
        await self._lock.acquire()
        await self.ensure_connection()
        if self.connection is None:
            raise ConnectionIsNotAccrued()
        return self.connection

    async def __aexit__(self, *_):
        self._lock.release()


class AsyncPoolConnectionWrapper[ConnType](BasePoolConnectionWrapper[ConnType]):
    def __init__(
        self, provider: BaseProvider[ConnType], pool_init_lock: asyncio.Lock
    ):
        self.provider = provider
        self.connection: ConnType | None = None
        self._pool_init_lock = pool_init_lock

    async def ensure_connection(self) -> None:
        if not self.provider.connections_pool:
            async with self._pool_init_lock:
                if not self.provider.connections_pool:
                    await self.provider.create_connection()

    async def __aenter__(self) -> ConnType:
        await self.ensure_connection()
        self.connection = await self.provider.connections_pool.acquire()
        if self.connection is None:
            raise ConnectionIsNotAccrued()
        return self.connection

    async def __aexit__(self, *_):
        await self.provider.connections_pool.release(self.connection)



from .core import BaseDBCore
from .providers import BaseSyncProvider


class SyncDBCore[Models](BaseDBCore[BaseSyncProvider, Models]):
    """
    Main sync db class.
    You can implement your queries to the database in the child class
    or use this class directly.
    """

    @classmethod
    def close_connections(cls) -> None:
        for provider in cls.dbs.values():
            provider.close_connection()

    def execute(self, query, args=()):
        return self.provider.execute(query, args)

    def create_tables(self) -> None:
        for signature in self.signatures.values():
            self.provider.executescript(
                self._prepare_create_table_query(signature)
            )

    def insert(self, objs, /) -> Models | list[Models]:
        obj_ids = self.provider.execute_insert_query(
            *self._prepare_insert_query(objs)
        )
        return self._assign_ids(objs, obj_ids)

    def fetchone(
        self,
        model,
        *,
        join=None,
        where=None,
        order_by=None,
        limit=None,
        offset=0,
    ):
        if data := self.provider.fetchone(
            *self._prepare_select_query(
                model.__name__, None, join, where, order_by, limit, offset
            )
        ):
            return self._convert_data(model, data, join)

    def fetchall(
        self,
        model,
        *,
        join=None,
        where=None,
        order_by=None,
        limit=None,
        offset=0,
    ):
        data = self.provider.fetchall(
            *self._prepare_select_query(
                model.__name__, None, join, where, order_by, limit, offset
            )
        )
        return (
            [self._convert_data(model, obj, join) for obj in data]
            if data
            else []
        )

    def save(self, obj) -> None:
        if (query := self._prepare_save_query(obj)) is not None:
            return self.execute(*query)

    def update(self, model, fields, *, where=None) -> None:
        self.execute(*self._prepare_update_query(model, fields, where))

    def delete(self, model, *, where) -> None:
        self.execute(*self._prepare_delete_query(model, where))

    def drop_table(self, model, /) -> None:
        """
        Drops table from db.
        :param model: model to drop.
        """
        self.execute(self._prepare_drop_table_query(model))



from __future__ import annotations

import threading
import typing as ty
from abc import ABC, abstractmethod

from ..exceptions import ConnectionIsNotAccrued
from .base import (
    BaseConnectionWrapper,
    BasePoolConnectionWrapper,
    BaseProvider,
    translate_exceptions,
)

if ty.TYPE_CHECKING:
    from .base import InsertQuery, Query, SelectQuery


class BaseSyncProvider[ConnType](BaseProvider[ConnType], ABC):
    """
    Base class for sync SQL providers.
    ConnType : type of connection instance.
    """

    @abstractmethod
    def ensure_connection(
        self,
    ) -> SyncConnectionWrapper[ConnType] | SyncPoolConnectionWrapper[ConnType]:
        raise NotImplementedError()

    @abstractmethod
    def _fetchone(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> tuple[ty.Any, ...] | None:
        raise NotImplementedError()

    @abstractmethod
    def _fetchall(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> list[tuple[ty.Any, ...]]:
        raise NotImplementedError()

    @translate_exceptions
    def execute(self, query: Query, args: ty.Sequence[ty.Any] = ()):
        args = tuple(self.adapt_value(arg) for arg in args)
        return self._execute(query, args)

    @translate_exceptions
    def execute_insert_query(
        self, query: InsertQuery, args: ty.Sequence[ty.Any]
    ) -> list[int]:
        args = tuple(self.adapt_value(arg) for arg in args)
        return [row[0] for row in self._fetchall(query, args)]

    @translate_exceptions
    def fetchone(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> tuple[ty.Any, ...] | None:
        args = tuple(self.adapt_value(arg) for arg in args)
        return self._fetchone(query, args)

    @translate_exceptions
    def fetchall(
        self, query: SelectQuery, args: ty.Sequence[ty.Any] = ()
    ) -> list[tuple[ty.Any, ...]]:
        args = tuple(self.adapt_value(arg) for arg in args)
        print(query)
        return self._fetchall(query, args)


class SyncConnectionWrapper[ConnType](BaseConnectionWrapper[ConnType]):
    def __init__(self, provider: BaseProvider[ConnType], lock: threading.Lock):
        self.provider = provider
        self.connection: ConnType | None = None
        self._lock = lock

    def ensure_connection(self) -> None:
        if not self.connection:
            self.provider.create_connection()
            self.connection = self.provider.connection

    def __enter__(self) -> ConnType:
        self._lock.acquire()
        self.ensure_connection()
        if self.connection is None:
            raise ConnectionIsNotAccrued()
        return self.connection

    def __exit__(self, *_):
        self._lock.release()


class SyncPoolConnectionWrapper[ConnType](BasePoolConnectionWrapper[ConnType]):
    def __init__(
        self, provider: BaseProvider[ConnType], pool_init_lock: threading.Lock
    ):
        self.provider = provider
        self.connection: ConnType | None = None
        self._pool_init_lock = pool_init_lock

    def ensure_connection(self) -> None:
        if not self.provider.connections_pool:
            with self._pool_init_lock:
                if not self.provider.connections_pool:
                    self.provider.create_connection()

    def __enter__(self) -> ConnType:
        self.ensure_connection()
        self.connection = self.provider.connections_pool.acquire()
        if self.connection is None:
            raise ConnectionIsNotAccrued()
        return self.connection

    def __exit__(self, *_):
        self.provider.connections_pool.release(self.connection)



import threading
import typing as ty

from ...exceptions import UniqueRequiredError

try:
    import sqlite3
except ModuleNotFoundError as err:
    raise RuntimeError(
        "You should install `sqlite3` backend to connect to this db. "
    ) from err

from ..base_sync import BaseSyncProvider, SyncConnectionWrapper


class Sqlite3Provider(BaseSyncProvider[sqlite3.Connection]):
    def __init__(self, db_path, **connection_kwargs) -> None:
        super().__init__(db_path, **connection_kwargs)
        self.connection = None
        self._lock = threading.Lock()

    def create_connection(self) -> None:
        if not self.connection:
            self.connection = sqlite3.connect(
                self.db_path, isolation_level=None
            )

    def close_connection(self) -> None:
        if self.connection:
            self.connection.close()
            self.connection = None

    def ensure_connection(self) -> SyncConnectionWrapper[sqlite3.Connection]:
        return SyncConnectionWrapper(self, self._lock)

    def _execute(self, query, args=()):
        with self.ensure_connection() as connection:
            return connection.execute(query, args)

    def executescript(self, query):
        with self.ensure_connection() as connection:
            return connection.executescript(query)

    def _fetchone(self, query, args=()) -> tuple[ty.Any] | None:
        if rows := self._fetchall(query, args):
            return rows[0]

    def _fetchall(self, query, args=()) -> list[tuple[ty.Any]]:
        with self.ensure_connection() as connection:
            return list(connection.execute(query, args).fetchall())

    @staticmethod
    def modify_db_path(db_path: str) -> str:
        return db_path.removeprefix("sqlite+sqlite3://")

    def _translate_exception(self, exception, query, params):
        if isinstance(exception, sqlite3.IntegrityError):
            if exception.sqlite_errorname == "SQLITE_CONSTRAINT_UNIQUE":
                text = exception.args[0]
                return UniqueRequiredError(
                    query,
                    params,
                    exception,
                    field_name=text[text.rfind(": ") + 2 :],
                )
        return super()._translate_exception(exception, query, params)



from __future__ import annotations

import typing as ty

if ty.TYPE_CHECKING:
    from .core import BaseDBCore
    from .joins import InnerJoin, LeftJoin, RightJoin
    from .models import Field
    from .operators import InvertedField, Operator
    from .providers import BaseSyncProvider


class SyncDBCore[Models](BaseDBCore[BaseSyncProvider, Models]):
    @classmethod
    def close_connections(cls) -> None: ...
    def execute(self, query, args=()): ...
    def create_tables(self) -> None: ...
    @ty.overload
    def insert[Model](self, objs: list[Model], /) -> list[Model]: ...
    @ty.overload
    def insert[Model](self, obj: Model, /) -> Model: ...
    @ty.overload
    def fetchone[Model](
        self,
        model: ty.Type[Model],
        *,
        join: None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> Model | None: ...
    @ty.overload
    def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: InnerJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model, JoinModel] | None: ...
    @ty.overload
    def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: LeftJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model, JoinModel | None] | None: ...
    @ty.overload
    def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: RightJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model | None, JoinModel] | None: ...
    @ty.overload
    def fetchall[Model](
        self,
        model: ty.Type[Model],
        *,
        join: None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[Model]: ...
    @ty.overload
    def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: InnerJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model, JoinModel]]: ...
    @ty.overload
    def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: LeftJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model, JoinModel | None]]: ...
    @ty.overload
    def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: RightJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model | None, JoinModel]]: ...
    def save(self, obj) -> None: ...
    def update(self, model, fields, *, where=None) -> None: ...
    def delete(self, model, *, where) -> None: ...
    def drop_table(self, model, /) -> None: ...
